require 'monitor'

class TaggedQueue
  def initialize
    @monitor = Monitor.new
    @any_cv = @monitor.new_cond
    @tag_cv = {}
    @queue = []
  end

  def enqueue(value, tag = value.class)
    @monitor.synchronize {
      @queue.push([tag, value])
      cv = @tag_cv[tag] and cv.signal
      @any_cv.signal
    }
    self
  end

  def dequeue(tag = nil)
    if tag.nil?
      @monitor.synchronize {
        @any_cv.wait_while { @queue.empty? }
        pair = @queue.shift
        return pair.last
      }
    else
      @monitor.synchronize {
        if not i = find_tag(tag)
          cv = (@tag_cv[tag] ||= @monitor.new_cond)
          cv.wait_until { i = find_tag(tag) }
        end
        pair = @queue.delete_at(i)
        return pair.last
      }
    end
  end

  private

  def find_tag(tag)
    @queue.each_with_index { |pair, i|
      return i if pair.first == tag
    }
    return nil
  end
end

if $0 == __FILE__
  eval DATA.read, nil, $0, __LINE__+4
end

__END__

queue = TaggedQueue.new
generators = {
  Integer => lambda { rand(10) },
  String  => lambda { (10 + rand(26)).to_s(36).upcase },
}
types = generators.keys

generators.each_pair { |type, func|
  Thread.start {
    puts "#{type} producer starts"
    loop {
      sleep rand(3)
      value = func.call
      puts "#{type} producer is enqueuing #{value}"
      queue.enqueue(value)
    }
  }
}

Thread.start {
  puts "General consumer starts"
  loop {
    sleep rand(3)
    value = queue.dequeue
    puts "General consumer got #{value}"
  }
}

types.each { |type|
  Thread.start {
    puts "#{type} consumer starts"
    loop {
      sleep rand(5)
      value = queue.dequeue(type)
      puts "#{type} consumer got #{value}"
    }
  }
}

sleep 30
